from .fx import Fx, fx

def _make_divider(divisor: str = "-",
                  count: int = 60) -> str:
    return divisor*count

def add(a: int | float | Fx,
        b: int | float | Fx) -> int | float | Fx:
    """
    Add two numbers
    :param a:
    :param b:
    :return:
    """
    if isinstance(a, float):
        a = fx(a)
    if isinstance(b, float):
        b = fx(b)
    return a + b

def subtract(a: int | float | Fx,
             b: int | float | Fx) -> int | float | Fx:
    """
    Subtract two numbers
    :param a:
    :param b:
    :return:
    """
    if isinstance(a, float):
        a = fx(a)
    if isinstance(b, float):
        b = fx(b)
    return a - b

def mul(a: int | float | Fx,
        b: int | float | Fx) -> int | float | Fx:
    """
    Multiply two numbers
    :param a:
    :param b:
    :return:
    """
    if isinstance(a, float):
        a = fx(a)
    if isinstance(b, float):
        b = fx(b)
    return a * b

def div(a: int | float | Fx,
        b: int | float | Fx) -> int | float | Fx:
    """
    Divide two numbers
    :param a:
    :param b:
    :return:
    """
    if isinstance(a, float):
        a = fx(a)
    if isinstance(b, float):
        b = fx(b)
    if b != 0:
        return a / b
    raise ZeroDivisionError


def sigma(collection: list[Fx] | list[int] | list[float]) -> Fx:
    return_result = 0
    for i in collection:
        return_result += i
    return return_result

def sigma_noiter(*numbers: Fx | int | float) -> Fx:
    return_result = Fx(0)
    for i in numbers:
        i = fx(i)
        return_result += i
    return return_result

def power(base: int | float | Fx,
        exp: int | float | Fx) -> int | float | Fx:
    if isinstance(base, (int, float)):
        base = fx(base)
    if isinstance(exp, (int, float)):
        exp = fx(exp)
    return base ** exp

if __name__ == "__main__":
    from tests import test_main
    test_main()



class UnexpectedTypeError(Exception):
    pass

class MissingArgument(Exception):
    pass

class ConstantError(Exception):
    def __init__(self, message: str = "Cannot modify a constant") -> None:
        super().__init__(message)


from .errors import UnexpectedTypeError

class Fx:
    def __init__(self, value: str | float | int) -> None:
        s = str(value)
        if '.' in s:
            whole, frac = s.split('.')
            self.scale: int = len(frac)
            self.value: int = int(whole + frac)
        else:
            self.scale: int = 0
            self.value: int = int(s)

        self._normalize()

    def is_zero(self) -> bool:
        return int(self.value) == 0

    def is_positive(self) -> bool:
        return int(self.value) > 0

    def is_negative(self) -> bool:
        return int(self.value) < 0

    def copy(self) -> Fx:
        new_copy = Fx.__new__(Fx)
        new_copy.value = self.value
        new_copy.scale = self.scale
        return new_copy

    def _normalize(self):
        """Remove trailing zeros from the fractional part"""
        if self.value == 0:
            self.scale = 0
            return

        while self.scale > 0 and self.value % 10 == 0:
            self.value //= 10
            self.scale -= 1

    def _align(self, other: Fx) -> tuple[int, int, int]:
        if not isinstance(other, Fx):
            raise UnexpectedTypeError(f"Unexpected type {type(other)}")

        if self.scale == other.scale:
            return self.value, other.value, self.scale
        if self.scale > other.scale:
            diff = self.scale - other.scale
            return self.value, other.value * (10 ** diff), self.scale
        diff = other.scale - self.scale
        return self.value * (10 ** diff), other.value, other.scale

    def __repr__(self):
        return f"Fx(value={self.value}, scale={self.scale})"

    #--- Conversions ---#
    def __float__(self):
        return self.value / (10 ** self.scale)

    def __int__(self):
        return int(float(self))

    def __str__(self):
        if self.scale == 0:
            return str(self.value)

        if self.value == 0:
            return "0"

        s = str(self.value).zfill(self.scale + 1)
        return s[:-self.scale] + '.' + s[-self.scale:]

    #--- Miscellaneous ---#
    def __neg__(self) -> Fx:
        result = Fx.__new__(Fx)
        result.value = -self.value
        result.scale = self.scale
        result._normalize()
        return result

    def __pos__(self) -> Fx:
        result = Fx.__new__(Fx)
        result.value = self.value
        result.scale = self.scale
        result._normalize()
        return result

    def __abs__(self) -> Fx:
        result = Fx.__new__(Fx)
        result.value = abs(self.value)
        result.scale = self.scale
        result._normalize()
        return result

    def __round__(self, n: int = 0) -> Fx:
        return Fx(round(float(self), n))

    def __index__(self) -> int:
        return int(self)

    def __format__(self, format_spec: str) -> str:
        return format(float(self), format_spec)

    #--- Comparisons ---#
    def __eq__(self, other):
        if isinstance(other, Fx):
            a, b, _ = self._align(other)
            return a == b
        if isinstance(other, (int, float)):
            return float(self) == other
        return False
    def __ne__(self, other):
        if isinstance(other, Fx):
            a, b, _ = self._align(other)
            return a != b
        if isinstance(other, (int, float)):
            return float(self) != other
        return False

    def __gt__(self, other):
        if isinstance(other, Fx):
            a, b, _ = self._align(other)
            return a > b
        if isinstance(other, (int, float)):
            return float(self) > other
        return False
    def __ge__(self, other):
        if isinstance(other, Fx):
            a, b, _ = self._align(other)
            return a >= b
        if isinstance(other, (int, float)):
            return float(self) >= other
        return False

    def __lt__(self, other):
        if isinstance(other, Fx):
            a, b, _ = self._align(other)
            return a < b
        if isinstance(other, (int, float)):
            return float(self) < other
        return False
    def __le__(self, other):
        if isinstance(other, Fx):
            a, b, _ = self._align(other)
            return a <= b
        if isinstance(other, (int, float)):
            return float(self) <= other
        return False

    def __bool__(self):
        return self.value != 0

    #--- Arithmetics --#
    def __add__(self, other: Fx | float | int) -> Fx:
        if isinstance(other, (int, float)):
            other = fx(other)
        a, b, scale = self._align(other)
        result_value = a + b
        result = Fx.__new__(Fx)
        result.value = result_value
        result.scale = scale
        result._normalize()
        return result

    def __sub__(self, other: Fx | float | int) -> Fx:
        if isinstance(other, (int, float)):
            other = fx(other)
        a, b, scale = self._align(other)
        result_value = a - b
        result = Fx.__new__(Fx)
        result.value = result_value
        result.scale = scale
        result._normalize()
        return result

    def __mul__(self, other: Fx | float | int) -> Fx:
        if isinstance(other, (int, float)):
            other = fx(other)
        result_value = self.value * other.value
        scale = self.scale + other.scale
        result = Fx.__new__(Fx)
        result.value = result_value
        result.scale = scale
        result._normalize()
        return result

    def __truediv__(self, other: Fx | float | int) -> Fx:
        if isinstance(other, (int, float)):
            other = fx(other)
        if other.value == 0:
            raise ZeroDivisionError

        a, b, _ = self._align(other)

        PRECISION = 10
        numerator = a * (10 ** PRECISION)

        result_value = numerator // b
        result_scale = PRECISION

        result = Fx.__new__(Fx)
        result.value = result_value
        result.scale = result_scale
        result._normalize()
        return result

    def __floordiv__(self, other: Fx | float | int) -> Fx:
        if isinstance(other, (int, float)):
            other = fx(other)
        if other.value == 0:
            raise ZeroDivisionError

        a, b, common_scale = self._align(other)

        numerator = a
        denominator = b * (10 ** common_scale)

        result_value = numerator // denominator

        result = Fx.__new__(Fx)
        result.value = result_value
        result.scale = 0
        result._normalize()
        return result

    def __pow__(self, other: Fx | float | int) -> Fx:
        if isinstance(other, (int, float)):
            other = fx(other)

        exponent = float(other)
        base = self

        if other.scale == 0 and other.value >= 0:
            exp_int = other.value
            if exp_int == 0:
                return Fx(1)
            result = Fx(1)
            for _ in range(exp_int):
                result *= base
            return result

        else:
            # NOTE: This uses floating-point arithmetic and may introduce errors.
            base_float = float(base)
            result_float = base_float ** exponent
            return Fx(result_float)

    def __mod__(self, other: Fx | float | int) -> Fx:
        if isinstance(other, (int, float)):
            other = fx(other)

        a, b, scale = self._align(other)

        result_value = a % b

        result = Fx.__new__(Fx)
        result.value = result_value
        result.scale = scale
        result._normalize()
        return result

    #--- Right hand ---#
    def __radd__(self, other: int | float) -> Fx:
        return self + other

    def __rsub__(self, other: int | float) -> Fx:
        return Fx(other) - self

    def __rmul__(self, other: int | float) -> Fx:
        return self * other

    def __rtruediv__(self, other: int | float) -> Fx:
        return Fx(other) / self

    def __rfloordiv__(self, other: int | float) -> Fx:
        return Fx(other) // self

    def __rmod__(self, other: int | float) -> Fx:
        return Fx(other) % self

    def __rpow__(self, other: int | float) -> Fx:
        return Fx(other) ** self

    #--- In-place operators ---#
    def __iadd__(self, other: Fx | int | float) -> Fx:
        result = self + other
        self.value = result.value
        self.scale = result.scale
        return self

    def __isub__(self, other: Fx | int | float) -> Fx:
        result = self - other
        self.value = result.value
        self.scale = result.scale
        return self

    def __imul__(self, other: Fx | int | float) -> Fx:
        result = self * other
        self.value = result.value
        self.scale = result.scale
        return self

    def __itruediv__(self, other: Fx | int | float) -> Fx:
        result = self / other
        self.value = result.value
        self.scale = result.scale
        return self

    def __ifloordiv__(self, other: Fx | int | float) -> Fx:
        result = self // other
        self.value = result.value
        self.scale = result.scale
        return self

    def __ipow__(self, other: Fx | int | float) -> Fx:
        result = self ** other
        self.value = result.value
        self.scale = result.scale
        return self

    def __imod__(self, other: Fx | int | float) -> Fx:
        result = self % other
        self.value = result.value
        self.scale = result.scale
        return self


def fx(value: str | int | float) -> Fx:
    """
    Converts the following `value` to fx, if possible.
    :param value:
    :return:
    """
    return Fx(value)

if __name__ == "__main__":
    from tests import test_fx
    test_fx()


from .main import _make_divider, add, subtract, mul, div, sigma, sigma_noiter
from .fx import Fx, fx
from .errors import UnexpectedTypeError, ConstantError
from .tests import test_main, test_fx
from .constant import Constant

__all__ = [
    "add",
    "subtract",
    "mul",
    "div",
    "sigma",
    "sigma_noiter",

    "Fx",
    "fx",

    "UnexpectedTypeError",
    "ConstantError",

    "test_main",
    "test_fx",
    "Constant",
]

__version__ = "0.2.0"
__author__ = "Razka Rizaldi"


from tinycolors import cprint, color, clib, cinput
from .main import _make_divider
from .fx import Fx, fx
from .errors import MissingArgument
from decimal import Decimal


def _create_test(title: str,
                 action: str,
                 expected: int | float,
                 actual: int | float | Fx,
                 _python: int | float | None = None,
                 _show_actual_python: bool = False,
                 _show_divisor: bool = True,
                 _decimal: Decimal | None = None,
                 _show_actual_decimal: bool = False) -> bool | tuple[bool, bool] | tuple[bool, bool, bool]:
    if _python is None and _show_actual_python:
        raise MissingArgument(f"Missing argument for '{_create_test.__name__}': '_python'")
    if _decimal is None and _show_actual_decimal:
        raise MissingArgument(f"Missing argument for '{_create_test.__name__}': '_decimal'")

    cprint(f"{title}", as_="bold yellow")
    cprint(f"{action} is...", as_="bold yellow")

    cprint(_make_divider('- ', 30), as_="bold white")
    cprint("Metrima says: ", as_="bold blue")
    print(f"{color.bg.black}{color.dim}{color.italic}{actual}{clib.reset}")

    cprint(_make_divider('- ', 30), as_="bold white")
    cprint("What we expected: ", as_="bold blue")
    cprint(f"{color.italic}{expected}", as_="bold green on black")

    cprint(_make_divider('- ', 30), as_="bold white")
    cprint("Metrima is..", as_="bold blue", end=" ")
    if actual == expected:
        cprint(f"{color.bold}CORRECT!", as_="italic green")
        cprint(f"{actual} is {expected}", as_="italic green")
    else:
        cprint(f"{color.bold}INCORRECT!", as_="italic red")
        cprint(f"{actual} is NOT {expected}", color="red")

    print()

    if not _show_actual_python and not _show_actual_decimal:
        cprint(_make_divider('='), as_="bold white") if _show_divisor else None
        return actual == expected

    if _show_actual_python:
        cprint("What Python says:", as_="bold cyan")
        _color = "bold red on black" if _python != expected else "bold green on black"
        cprint(f"{color.italic}{_python}", as_=_color)
        print()

    if _show_actual_decimal:
        cprint("What Decimal says:", as_="bold magenta")
        _decimal_float = float(_decimal)
        _color = "bold red on black" if _decimal_float != expected else "bold green on black"
        cprint(f"{color.italic}{_decimal}", as_=_color)

    cprint(_make_divider('='), as_="bold white") if _show_divisor else None

    if _show_actual_python and _show_actual_decimal:
        return actual == expected, _python == expected, _decimal_float == expected
    elif _show_actual_python:
        return actual == expected, _python == expected
    else:
        return actual == expected


def test_main() -> None:
    """
    Test demo for Metrima
    :return:
    """
    cprint(_make_divider('='), as_="bold white")
    cprint("             Metrima v0.1.0 Testing Grounds             ", as_="bold white")
    cprint("        Metrima Demo, Metrima VS Python VS Decimal      ", as_="bold white")
    cprint(_make_divider('='), as_="bold white")
    total = 0
    passed = 0
    pytotal = 0
    pypassed = 0
    dectotal = 0
    decpassed = 0

    # -- ADDITION -- #
    result = _create_test("Addition", "1.432 + 5.1234", 6.5554,
                          fx(1.432) + fx(5.1234),
                          1.432 + 5.1234, True,
                          True,
                          Decimal('1.432') + Decimal('5.1234'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- SUBTRACTION -- #
    result = _create_test("Subtraction", "10.0 - 2.45", 7.55,
                          fx(10.0) - fx(2.45),
                          10.0 - 2.45, True,
                          True,
                          Decimal('10.0') - Decimal('2.45'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- MULTIPLICATION -- #
    result = _create_test("Multiplication", "7.2 * 41.512", 298.8864,
                          fx(7.2) * fx(41.512),
                          7.2 * 41.512, True,
                          True,
                          Decimal('7.2') * Decimal('41.512'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- DIVISION -- #
    result = _create_test("Division", "3.6 / 1.2", 3.0,
                          fx(3.6) / fx(1.2),
                          3.6 / 1.2, True,
                          True,
                          Decimal('3.6') / Decimal('1.2'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- POWER/EXPONENTIATION -- #
    result = _create_test("Power", "2.5 ** 3", 15.625,
                          fx(2.5) ** fx(3),
                          2.5 ** 3, True,
                          True,
                          Decimal('2.5') ** Decimal('3'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- MODULO -- #
    result = _create_test("Modulo", "10.5 % 3.2", 0.9,
                          fx(10.5) % fx(3.2),
                          10.5 % 3.2, True,
                          True,
                          Decimal('10.5') % Decimal('3.2'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- FLOOR DIVISION -- #
    result = _create_test("Floor Division", "17.8 // 4.0", 4.0,
                          fx(17.8) // fx(4.0),
                          17.8 // 4.0, True,
                          True,
                          Decimal('17.8') // Decimal('4.0'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- NEGATION -- #
    result = _create_test("Negation", "-(-5.75)", 5.75,
                          -fx(-5.75),
                          -(-5.75), True,
                          True,
                          -(-Decimal('5.75')), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- ABSOLUTE VALUE -- #
    result = _create_test("Absolute Value", "abs(-12.345)", 12.345,
                          abs(fx(-12.345)),
                          abs(-12.345), True,
                          True,
                          abs(Decimal('-12.345')), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- COMPLEX EXPRESSION -- #
    result = _create_test("Complex Expression", "(2.5 + 3.5) * 4 - 10 / 2", 19.0,
                          (fx(2.5) + fx(3.5)) * fx(4) - fx(10) / fx(2),
                          (2.5 + 3.5) * 4 - 10 / 2, True,
                          True,
                          (Decimal('2.5') + Decimal('3.5')) * Decimal('4') - Decimal('10') / Decimal('2'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- CHAINED OPERATIONS -- #
    result = _create_test("Chained Addition", "1.1 + 2.2 + 3.3 + 4.4", 11.0,
                          fx(1.1) + fx(2.2) + fx(3.3) + fx(4.4),
                          1.1 + 2.2 + 3.3 + 4.4, True,
                          True,
                          Decimal('1.1') + Decimal('2.2') + Decimal('3.3') + Decimal('4.4'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # -- MIXED OPERATIONS -- #
    result = _create_test("Mixed Operations", "5 * 3 + 12 / 4 - 2", 16.0,
                          fx(5) * fx(3) + fx(12) / fx(4) - fx(2),
                          5 * 3 + 12 / 4 - 2, True, False,
                          Decimal('5') * Decimal('3') + Decimal('12') / Decimal('4') - Decimal('2'), True)
    total += 1
    pytotal += 1
    dectotal += 1
    passed += result[0]
    pypassed += result[1]
    decpassed += result[2]

    # Summary
    cprint("\n" + _make_divider('='), as_="bold white")
    cprint("                    TESTING COMPLETE!                    ", as_="bold green")
    cprint(_make_divider('='), as_="bold white")

    text_color = "bold cyan" if passed != total else "bold blue"
    cprint(f"\nMetrima Results: {passed}/{total} tests passed", as_=text_color)
    metrima_percentage = (passed / total * 100) if total > 0 else 0
    metrima_color = "bold green" if passed == total else "bold yellow" if passed > total * 0.5 else "bold red"
    cprint(f"Success Rate: {metrima_percentage:.1f}%", as_=metrima_color)

    text_color = "bold cyan" if pypassed != total else "bold magenta"
    cprint(f"\nPython Results: {pypassed}/{pytotal} tests passed", as_=text_color)
    python_percentage = (pypassed / pytotal * 100) if pytotal > 0 else 0
    python_color = "bold green" if pypassed == pytotal else "bold yellow" if pypassed > pytotal * 0.5 else "bold red"
    cprint(f"Success Rate: {python_percentage:.1f}%", as_=python_color)

    text_color = "bold cyan" if decpassed != total else "bold magenta"
    cprint(f"\nDecimal Results: {decpassed}/{dectotal} tests passed", as_=text_color)
    decimal_percentage = (decpassed / dectotal * 100) if dectotal > 0 else 0
    decimal_color = "bold green" if decpassed == dectotal else "bold yellow" if decpassed > dectotal * 0.5 else "bold red"
    cprint(f"Success Rate: {decimal_percentage:.1f}%", as_=decimal_color)

    cprint("\n" + _make_divider('='), as_="bold white")

def test_fx():
    cprint(_make_divider('='), as_="bold white")
    cprint("             Metrima v0.1.0 Testing Grounds             ", as_="bold white")
    cprint(f"       Metrima's {color.bg.black}Fx{clib.reset} testing      ", as_="bold white")
    cprint(_make_divider('='), as_="bold white")
    print()

    test_count = 0
    pass_count = 0

    cprint("Fx class testing", as_="bold white on black")
    print()


    pass_local = 0
    total_local = 6

    cprint("repr(Fx()) and str(Fx())", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1, expected2, expected3 = 'Fx(value=0, scale=0)', 'Fx(value=0, scale=0)', 'Fx(value=0, scale=0)'
    actual1, actual2, actual3 = repr(Fx("0")), repr(Fx(0)), repr(Fx(0.0))

    cprint(f"repr(Fx(\"0\")) = {actual1}", "cyan")
    cprint(f"repr(Fx(0)) = {actual2}", "cyan")
    cprint(f"repr(Fx(0.0)) = {actual3}", "cyan")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1

    expected4, expected5, expected6 = "3.14", "42", "0.5"
    actual4, actual5, actual6 = str(Fx("3.14")), str(Fx(42)), str(Fx(0.5))

    cprint(f"str(Fx(\"3.14\")) = {actual4}", "cyan")
    cprint(f"str(Fx(42)) = {actual5}", "cyan")
    cprint(f"str(Fx(0.5)) = {actual6}", "cyan")
    cprint('-'*40, as_="bold white")

    if actual4 == expected4:
        pass_local += 1
    if actual5 == expected5:
        pass_local += 1
    if actual6 == expected6:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 3

    cprint("Addition: 1.432 + 5.1234", as_="bold yellow")
    cprint('-' * 40, as_="bold white")

    expected1 = str(Fx("6.5554"))
    actual1 = str(Fx("1.432") + Fx("5.1234"))
    cprint(f'Fx("1.432") + Fx("5.1234") = {actual1}', "cyan")

    expected2 = str(Fx("5.5"))
    actual2 = str(Fx(2) + Fx(3.5))
    cprint(f'Fx(2) + Fx(3.5) = {actual2}', "cyan")

    expected3 = str(Fx("0.019"))
    actual3 = str(Fx(0.01) + Fx(0.009))
    cprint(f'Fx(0.01) + Fx(0.009) = {actual3}', "cyan")

    cprint('-' * 40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 3

    cprint("Subtraction: 5.5 - 2.2", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = str(Fx("3.3"))
    actual1 = str(Fx("5.5") - Fx("2.2"))
    cprint(f'Fx("5.5") - Fx("2.2") = {actual1}', "cyan")

    expected2 = str(Fx("7"))
    actual2 = str(Fx(10) - 3)
    cprint(f'Fx(10) - 3 = {actual2}', "cyan")

    expected3 = str(Fx("4.5"))
    actual3 = str(7 - Fx(2.5))
    cprint(f'7 - Fx(2.5) = {actual3}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 3

    cprint("Multiplication: 1.5 * 2", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = str(Fx("3.0"))
    actual1 = str(Fx("1.5") * Fx("2"))
    cprint(f'Fx("1.5") * Fx("2") = {actual1}', "cyan")

    expected2 = str(Fx("7.5"))
    actual2 = str(Fx(3) * 2.5)
    cprint(f'Fx(3) * 2.5 = {actual2}', "cyan")

    expected3 = str(Fx("8.4"))
    actual3 = str(2 * Fx(4.2))
    cprint(f'2 * Fx(4.2) = {actual3}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 3

    cprint("Division: 3.6 / 1.2", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = str(Fx("3.0"))
    actual1 = str(Fx("3.6") / Fx("1.2"))
    cprint(f'Fx("3.6") / Fx("1.2") = {actual1}', "cyan")

    expected2 = str(Fx("2.5"))
    actual2 = str(Fx(5) / 2)
    cprint(f'Fx(5) / 2 = {actual2}', "cyan")

    expected3 = str(Fx("3.0"))
    actual3 = str(12 / Fx(4))
    cprint(f'12 / Fx(4) = {actual3}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 3

    cprint("Floor Division: 2.5 // 1.2", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = str(Fx("2.0"))
    actual1 = str(Fx("2.5") // Fx("1.2"))
    cprint(f'Fx("2.5") // Fx("1.2") = {actual1}', "cyan")

    expected2 = str(Fx("2.0"))
    actual2 = str(Fx(7) // 3)
    cprint(f'Fx(7) // 3 = {actual2}', "cyan")

    expected3 = str(Fx("2.0"))
    actual3 = str(10 // Fx(4))
    cprint(f'10 // Fx(4) = {actual3}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 4

    cprint("Comparisons", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = True
    actual1 = Fx("2.0") == Fx("2")
    cprint(f'Fx("2.0") == Fx("2") = {actual1}', "cyan")

    expected2 = False
    actual2 = Fx(3) != 3
    cprint(f'Fx(3) != 3 = {actual2}', "cyan")

    expected3 = True
    actual3 = Fx("1.5") > Fx("1.4")
    cprint(f'Fx("1.5") > Fx("1.4") = {actual3}', "cyan")

    expected4 = True
    actual4 = Fx("1.2") < 2
    cprint(f'Fx("1.2") < 2 = {actual4}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1
    if actual4 == expected4:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 3

    cprint("Modulo: 5.5 % 2", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = str(Fx("1.5"))
    actual1 = str(Fx("5.5") % Fx("2"))
    cprint(f'Fx("5.5") % Fx("2") = {actual1}', "cyan")

    expected2 = str(Fx("1"))
    actual2 = str(Fx(7) % 3)
    cprint(f'Fx(7) % 3 = {actual2}', "cyan")

    expected3 = str(Fx("0"))
    actual3 = str(10 % Fx(2.5))
    cprint(f'10 % Fx(2.5) = {actual3}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 3

    cprint("Power: 2 ** 3", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = str(Fx("8.0"))
    actual1 = str(Fx("2") ** Fx("3"))
    cprint(f'Fx("2") ** Fx("3") = {actual1}', "cyan")

    expected2 = str(Fx("9.0"))
    actual2 = str(Fx(3) ** 2)
    cprint(f'Fx(3) ** 2 = {actual2}', "cyan")

    expected3 = str(Fx("32.0"))
    actual3 = str(2 ** Fx(5))
    cprint(f'2 ** Fx(5) = {actual3}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 5

    cprint("Type Conversions", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = 3.14
    actual1 = float(Fx("3.14"))
    cprint(f'float(Fx("3.14")) = {actual1}', "cyan")

    expected2 = 42
    actual2 = int(Fx("42.7"))
    cprint(f'int(Fx("42.7")) = {actual2}', "cyan")

    expected3 = "5.5"
    actual3 = str(Fx(5.5))
    cprint(f'str(Fx(5.5)) = {actual3}', "cyan")

    expected4 = True
    actual4 = bool(Fx("3.14"))
    cprint(f'bool(Fx("3.14")) = {actual4}', "cyan")

    expected5 = False
    actual5 = bool(Fx(0))
    cprint(f'bool(Fx(0)) = {actual5}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1
    if actual4 == expected4:
        pass_local += 1
    if actual5 == expected5:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 4

    cprint("Unary Operations", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    expected1 = str(Fx("-3.14"))
    actual1 = str(-Fx("3.14"))
    cprint(f'-Fx("3.14") = {actual1}', "cyan")

    expected2 = str(Fx("5.5"))
    actual2 = str(+Fx("5.5"))
    cprint(f'+Fx("5.5") = {actual2}', "cyan")

    expected3 = str(Fx("3.14"))
    actual3 = str(abs(Fx("-3.14")))
    cprint(f'abs(Fx("-3.14")) = {actual3}', "cyan")

    expected4 = str(Fx("3.1"))
    actual4 = str(round(Fx("3.14159"), 1))
    cprint(f'round(Fx("3.14159"), 1) = {actual4}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1
    if actual4 == expected4:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    pass_local = 0
    total_local = 4

    cprint("In-place Operations", as_="bold yellow")
    cprint('-'*40, as_="bold white")

    x1 = Fx("5")
    x1 += Fx("3")
    expected1 = str(Fx("8"))
    actual1 = str(x1)
    cprint(f'x = Fx("5"); x += Fx("3"); x = {actual1}', "cyan")

    x2 = Fx("10")
    x2 -= 3
    expected2 = str(Fx("7"))
    actual2 = str(x2)
    cprint(f'x = Fx("10"); x -= 3; x = {actual2}', "cyan")

    x3 = Fx("4")
    x3 *= 2.5
    expected3 = str(Fx("10.0"))
    actual3 = str(x3)
    cprint(f'x = Fx("4"); x *= 2.5; x = {actual3}', "cyan")

    x4 = Fx("9")
    x4 //= 2
    expected4 = str(Fx("4.0"))
    actual4 = str(x4)
    cprint(f'x = Fx("9"); x //= 2; x = {actual4}', "cyan")

    cprint('-'*40, as_="bold white")

    if actual1 == expected1:
        pass_local += 1
    if actual2 == expected2:
        pass_local += 1
    if actual3 == expected3:
        pass_local += 1
    if actual4 == expected4:
        pass_local += 1

    test_count += 1
    if pass_local == total_local:
        pass_count += 1
        cprint(f"Pass, {pass_local} out of {total_local}", as_="bold green")
    else:
        cprint(f"Fail, {pass_local} out of {total_local}", as_="bold red")

    print()


    cprint("="*40, as_="bold white")
    cprint(f"TOTAL: {pass_count} out of {test_count} test containers passed", as_="bold white on black")
    if pass_count == test_count:
        cprint("ALL TEST CONTAINERS PASS", as_="bold green")
    else:
        cprint(f"{test_count - pass_count} test containers failed", as_="bold red")
    cprint("="*40, as_="bold white")

def menu():
    cprint(_make_divider('='), as_="bold white")
    cprint("             Metrima v0.1.0 Testing Grounds              ", as_="bold white")
    cprint(" Welcome to Metrima! A small, comprehensive math library.", as_="bold white")
    cprint(_make_divider('='), as_="bold white")

    print()
    cprint(f'{color.italic}1. Competition Demo', as_="bold bright red")
    cprint('2. Fx Class Testing', as_="italic blue")
    user_choice = cinput(f"{color.italic}Which test would you like to run? ", as_="dim default")
    print()

    _make_divider()
    if user_choice in ["1", "competition", "c", "comp", "main"]:
        test_main()
    elif user_choice in ["2", "fx"]:
        test_fx()
    elif user_choice in ["quit", "exit", "q", "x"]:
        exit(0)
    else:
        cprint('x Invalid choice', as_="bold red")

if __name__ == "__main__":
    menu()


from .errors import ConstantError


class Constant:
    """
    An immutable container for a single constant value.
    Attempts to change any attribute after initialization will fail.
    """
    def __init__(self, value: int) -> None:
        object.__setattr__(self, 'value', value)

    def __setattr__(self, name, value) -> None:
        """Prevents modification of any attribute after __init__."""
        raise ConstantError()

    def __delattr__(self, name) -> None:
        raise ConstantError()

    def __repr__(self) -> str:
        return f"Constant({self.value!r})"

    def __str__(self) -> str:
        return str(self.value)